#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "as.h"
#include "table.h"
#include "lookup.h"
#include "regs.h"
#include "funcs.h"

extern int Size;
extern char Label[];
extern char Line[];
extern char *Optr;
extern long Pc;
extern int Lflag;
extern int Fwdsize;
extern long Old_pc;
extern int P_total;
extern int P_force;
char *next_str();

struct regs iregs[] = { /* pre-defined internal register names */
{"CACR", CREG,   CACR},
{"cacr", CREG,   CACR},
{"VBR",  CREG,   VBR},
{"vbr",  CREG,   VBR},
{"ACR0",	CREG,   ACR0},
{"acr0",	CREG,   ACR0},
{"ACR1",	CREG,   ACR1},
{"acr1",	CREG,   ACR1},
{"ROMBAR", CREG, ROMBAR},
{"rombar", CREG, ROMBAR},
{"RAMBAR", CREG, RAMBAR},
{"rambar", CREG, RAMBAR},
{"RAMBAR0", CREG, RAMBAR},
{"rambar0", CREG, RAMBAR},
{"SRAMBAR", CREG, RAMBAR},
{"srambar", CREG, RAMBAR},
{"MBAR",	CREG, 	MBAR},
{"mbar",	CREG, 	MBAR},
{"A0",   AREG,   0},
{"a0",   AREG,   0},
{"A1",   AREG,   1},
{"a1",   AREG,   1},
{"A2",   AREG,   2},
{"a2",   AREG,   2},
{"A3",   AREG,   3},
{"a3",   AREG,   3},
{"A4",   AREG,   4},
{"a4",   AREG,   4},
{"A5",   AREG,   5},
{"a5",   AREG,   5},
{"A6",   AREG,   6},
{"a6",   AREG,   6},
{"A7",   AREG,   7},
{"SP",   AREG,   7},
{"a7",   AREG,   7},
{"sp",   AREG,   7},
{"D0",   DREG,   0},
{"d0",   DREG,   0},
{"D1",   DREG,   1},
{"d1",   DREG,   1},
{"D2",   DREG,   2},
{"d2",   DREG,   2},
{"D3",   DREG,   3},
{"d3",   DREG,   3},
{"D4",   DREG,   4},
{"d4",   DREG,   4},
{"D5",   DREG,   5},
{"d5",   DREG,   5},
{"D6",   DREG,   6},
{"d6",   DREG,   6},
{"D7",   DREG,   7},
{"d7",   DREG,   7},
{"CCR",  SREG,   CCR},
{"ccr",  SREG,   CCR},
{"SR",   SREG,   SR},
{"sr",   SREG,   SR},
{"PC",   PC,     0},
{"pc",   PC,     0}
};
#define NIREGS  (sizeof(iregs)/sizeof(iregs[0]))

struct opts opt_list[] = {
{	"l",            o_list},
{	"list",         o_list},
{	"nol",          o_nolist},
{	"nolist",       o_nolist},
{	"brl",          o_null},
{	0,              0}
	};

/* range checking categories */
#define UBYTE   0       /* unsigned byte */
#define SBYTE   1       /* signed byte */
#define XBYTE   2       /* extended byte */
#define UWORD   3       /* unsigned word */
#define SWORD   4       /* signed word */
#define XWORD   5       /* extended word */
#define QUK     6       /* quick value 1-8 */
#define LOW3    7       /* 3 bit field */
#define LOW4    8       /* 4 */
#define LOW5    9       /* 5 */
#define LOW6    10	    /* 6 */
#define LOW7    11      /* 7 */
#define LOW7S   12      /* 7, signed */

struct ranges {
	long     r_min;  /* minimum value allowed */
	long     r_max;  /* max. allowed */
	long    r_mask; /* return value mask */
	};

struct ranges ckrange[] = {
  {	0,      MAXUBYTE,       0x00FF},
  {	MINBYTE,MAXBYTE,        0x00FF},
  {	MINBYTE,MAXUBYTE,       0x00FF},
  {	0,      MAXUWORD,       0xFFFF},
  {	MINWORD,MAXWORD,        0xFFFF},
  {	MINWORD,MAXUWORD,       0xFFFF},
  {	1,      8,              0x0007},
  {	0,      7,              0x0007},
  {	0,      15,             0x000F},
  {	0,      31,             0x001F},
  {	0,	63,		0x003F},
  {	0,      127,            0x007F},
  {	-64,    63,             0x007F},
	};

/*
 *      localinit --- machine specific initialization
 */
void
localinit(void)
{
	register struct regs *r;

	/* install register names in symbol table */
	for(r=iregs; r<&iregs[NIREGS]; r++)
		install(r->r_name,r->r_value,r->r_type);
}

/*
 *      do_op --- process mnemonic
 */
void
do_op(int m_class,int op,int op2)
/*int m_class;	    mnemonic class */
/*int op;          base opcode */
/*int op2;         second opcode (e.g. F-lines) */
{
	extern struct ea Eas[];
	register struct ea *ea1 = &Eas[0];
	register struct ea *ea2 = &Eas[1];
	register long tmp;
	register long dist;      /* distance on branches */
	register long rlist;     /* bit map of register list */
	char    *ptmp;

	switch(m_class){
	case INH:               /* inherent */
		eword(op);
		break;

	case RXRY:              /* Rx and Ry, no size */
		eword( op + (ea2->reg<<9) + ea1->reg );
		break;

	case RXRYS:             /* Rx and Ry, sized */
		eword( op + (ea2->reg<<9) + size76() + ea1->reg );
		break;

	case RXRYR:             /* Rx and Ry, reversed */
		eword( op + (ea1->reg<<9) + ea2->reg );
		break;

	case EAREG:             /* ea to register */
		eword( op + (ea2->reg<<9) + modreg(ea1) );
		finish(ea1);
		break;

	case EAREGS:            /* ea to register, sized */
		eword( op + (ea2->reg<<9) + size76() + modreg(ea1));
		finish(ea1);
		break;

	case REGEA:             /* register to ea */
		eword( op + (ea1->reg<<9) + modreg(ea2));
		finish(ea2);
		break;

	case REGEAS:            /* register to ea, signed */
		eword( op + (ea1->reg<<9) + size76() + modreg(ea2));
		finish(ea2);
		break;

   case IMMEAS:            /* immediate to ea, sized */
		eword( op + size76() + modreg(ea2) );
		if(Size==L)
			elong(ea1->constant);
		else
			eword(ea1->constant);
		finish(ea2);
		break;

	case QUKEA:             /* quick immediate to ea */
		if( (Size&B) && ea2->type==AN )
			error("Byte not allowed to address reg.");
		eword( op + (check(ea1->constant,QUK)<<9) + size76() + modreg(ea2) );
		finish(ea2);
		break;

	case IMMB:              /* immediate byte */
		eword(op);
		eword(check(ea1->constant,UBYTE));
		break;

	case IMMW:              /* immediate word */
		eword(op);
		eword(check(ea1->constant,XWORD));
		break;

	case IMMWS:             /* immediate word, signed */
		eword(op);
		eword(check(ea1->constant,SWORD));
		break;

	case IMM3:              /* immediate value, 3 bits */
		eword( op + check(ea1->constant,LOW3) );
		break;

	case IMM4:              /* immediate value, 4 bits */
		eword( op + check(ea1->constant,LOW4) );
		break;

	case RSHIFT:            /* register shift */
		eword( op + (ea1->reg<<9) + size76() + ea2->reg );
		break;

	case QSHIFT:            /* immediate (quick fmt) shift */
		eword( op + (check(ea1->constant,QUK)<<9) + size76() + ea2->reg);
		break;

	case EA:                /* ea */
		eword( op + modreg(ea1) );
		finish(ea1);
		break;

	case EAREV:             /* ea, reversed */
		eword( op + modreg(ea2) );
		finish(ea2);
		break;

	case EAS:               /* ea, sized */
		eword( op + size76() + modreg(ea1) );
		finish(ea1);
		break;

	case BCC:               /* conditional branches */
		dist = ea1->constant - (Pc+2);
		if(Size==U && ea1->force )
			Size = Fwdsize;
		if( dist<MINWORD || dist>MAXWORD )
			error("Error: displacement larger that 16-bits");
		if(Size==U){
			if( dist<MINBYTE || dist>MAXBYTE )
				Size = L;
			else
				Size = W;
      }
		switch(Size){
      /* 8-bit displacement */
		case W:
			if( dist==0 || dist== -1)
				error("Bad branch destination");
			eword( op + check(dist,SBYTE) );
			break;
      /* 16-bit displacement */
		case L:
			eword(op);
			eword(check(dist,SWORD));
			break;
   	}
      break;
	case BIT:               /* single bit manipulation */
		eword( op + modreg(ea2) );
		eword( check(ea1->constant,LOW5) );
		finish(ea2);
		break;

	case CHK:               /* chk */
		if( Size != L )
			op |= 0x0080;
		eword( op + (ea2->reg<<9) + modreg(ea1) );
		finish(ea1);
		break;

	case CHK2:              /* chk2,cmp2 */
		eword( op + size109() + modreg(ea1) );
		eword( op2 + adreg(ea2) );
		finish(ea1);
		break;

	case DBCC:              /* dbcc */
		dist = ea2->constant - (Old_pc+2);
		eword(op + ea1->reg );
		eword(check(dist,SWORD));
		break;

	case MULDIV:            /* multiplies and divides */
		eword( op + modreg(ea1) );
		if( ea2->type == DN )
			ea2->reg2 = ea2->reg;
		eword( op2 + (ea2->reg2<<12) + ea2->reg);
		finish(ea1);
		break;

	case REG:               /* register */
		eword(op + ea1->reg);
		break;

	case MOVEU:             /* move to/from USP */
		if( ea1->type == CN ){
			eword( op + ea2->reg);
			}
		else
			eword( op + ea1->reg);
      break;

	case REGIMM:            /* register with immediate */
		eword(op + ea1->reg);
		if( Size == L)
			elong(ea2->constant);
		else
			eword(ea2->constant);
		break;

	case MOVE:              /* move */
		tmp = modreg(ea2);
		tmp = ((tmp&7)<<3) + ((tmp>>3)&7);      /* reverse modreg */
		eword( op + (tmp<<6) + modreg(ea1) );
		finish(ea1);
		finish(ea2);
		break;

	case MOVEC:             /* movec */
		eword(op);
		if( ea1->type == CN )
			eword( adreg(ea2) + ea1->reg );
		else
			eword( adreg(ea1) + ea2->reg );
		break;

	case MOVEQ:             /* move quick */
		eword(op + (ea2->reg<<9) + check(ea1->constant,SBYTE) );
		break;

	case MOVEMO:            /* movem register to memory */
		if(ea1->type == IMMED)
			rlist = ea1->constant;
		else
			rlist = getrlist(ea1);
		if( ea2->type == PREDEC )
			rlist = reverse(rlist)>>16;	/* assumes int=32bits */
		eword( op + modreg(ea2) );
		eword(rlist);
		finish(ea2);
		break;

	case MOVEMI:            /* movem memory to register */
		if(ea2->type == IMMED)
			rlist = ea2->constant;
		else
			rlist = getrlist(ea2);
		/* predec not possible here */
		eword(op + modreg(ea1));
		eword(rlist);
		finish(ea1);
		break;

	case MOVEPO:            /* movep out */
		if(ea2->itype != D16AN )
			error("Simple indexing only");
		eword(op + (ea1->reg<<9) + ea2->reg);   /* NOT modreg(ea2) */
		finish(ea2);
		break;

	case MOVEPI:            /* movep in */
		if(ea1->itype != D16AN )
			error("Simple indexing only");
		eword(op + (ea2->reg<<9) + ea1->reg);   /* NOT modreg(ea1) */
		finish(ea1);
		break;

	case MOVES:             /* moves */
		op += size76();
		if( ea1->type == DN || ea1->type == AN ){
			eword( op + modreg(ea2) );
			eword( op2 + adreg(ea1) );
			finish(ea2);
			}
		else{
			eword( op + modreg(ea1) );
			eword( op2 + adreg(ea2) );
			finish(ea1);
			}
		break;

	case TRAPCC:            /* trapcc */
		eword(op + (Size==W ? 2 : 3));
		if( Size == L )
			elong(ea1->constant);
		else
			eword(check(ea1->constant,UWORD));
		break;

	case WDEBUG:              /* wdebug */
		eword( op + modreg(ea1) );
		eword( op2 );
		finish(ea1);
		break;

	case EQU:               /* equate */
		if(*Label){
			install(Label,ea1->constant,SYM);
			Old_pc = ea1->constant;    /* override normal */
			P_force++;
			}
		else
			error("EQU requires label");
		break;

	case DC:                /* define constants */
		while(ea1->type != EMPTY ){
			if( ea1->type == EXPR){
				switch(Size){
				case B: emit((WORD)ea1->constant); break;
				case U:
				case W: eword(ea1->constant); break;
				case L: elong(ea1->constant); break;
					}
				}
			else{   /* must be STRING */
				ptmp = (char *)ea1->constant;
				switch(Size){
				case B: while(*ptmp)emit(*ptmp++); break;
				case U:
				case W: while(*ptmp)eword(*ptmp++); break;
				case L: while(*ptmp)elong(*ptmp++); break;
					}
				free(&ea1->constant);
				}
			ea1++;
			}
		break;

	case OPT:               /* assembler options */
		do_opt(opt_list);
		break;

	case RORG:              /* relocatable org */
		warn("ORG assumed");
	case ORG:               /* origin */
      if (ea1->constant & 0x1)	/* word alignment */
      	ea1->constant -= 1;
		Old_pc = Pc = ea1->constant;
		f_record();     /* flush out any bytes */
		P_force++;
		break;

	case DS:                /* define storage */
		switch(Size){
		case B: tmp = 1; break;
		case U:
		case W: tmp = 2; break;
		case L: tmp = 4; break;
		}
		if(ea1->constant){
			tmp *= ea1->constant;
			while(tmp--)
				emit(0);
			break;
			}
		ea1->constant = tmp;       /* #items was zero, implies align */
		/* fall through to align code */
///////////////////////////////////////////////////////////////////////////////
/* under construction */
	case ALIGN:                     /* align Pc to boundary */
/*		if(ea1->constant && (tmp == Pc % ea1->constant)){
			tmp = ea1->constant - tmp;
			while(tmp--)
				emit(0);
			}
*/		break;
////////////////////////////////////////////////////////////////////////////////

	case INCLUDE:           /* process include file */
		make_pass(next_str());
		P_force = P_total = 0;
		Line[0] = '\n';
		break;

	case NULL_OP:           /* ignored mnemonic */
		break;
	default:
		fatal("Error in Mnemonic table");
		}
}

/*
 *      check --- verify that constant is within bounds
 */
long
check(long constant,int type)
{
	char    cbuf[100];
	long     lo,hi;
	extern int Pass;

	if( Pass==2 ){
		lo = ckrange[type].r_min;
		hi = ckrange[type].r_max;
		if( constant<lo || constant>hi ){
			sprintf(cbuf,"Value not in range: [%d,%d]",lo,hi);
			warn(cbuf);
			}
		}
	return(constant&ckrange[type].r_mask);
}

/*
 *      size76 --- compute opcode bits 7 and 6 from Size specifier
 */
int
size76(void)
{
	int s=0;

	switch(Size){
	case B: s = 0x00; break;
	case W: s = 0x40; break;
	case U:
	case L: s = 0x80; break;
	default:
		fatal("Bad size in size76");
		}
	return(s);
}

/*
 *      size109 --- compute opcode bits 10 and 9 from size specifier
 */
int
size109(void)
{
	int s=0;

	switch(Size){
	case B: s = 0x000; break;
	case W:
	case U: s = 0x200; break;
	case L: s = 0x400; break;
	default:
		fatal("Bad size in size109");
		}
	return(s);
}

/*
 *      size109b --- compute opcode bits 10 and 9 from size specifier (alternate)
 */
int
size109b(void)
{
	int s = 0;

	switch(Size){
	case B: s = 0x200; break;
	case W:
	case U: s = 0x400; break;
	case L: s = 0x600; break;
	default:
		fatal("Bad size in size109b");
		}
	return(s);
}

/*
 *      adreg --- return 4 bit encoding for An or Dn (shift left by 12)
 */
int
adreg(struct ea* e)
{
	int r = (int)e->reg;

	if( e->type == AN )
		r += 8;
	return( r<<12);
}

/*
 *	modreg --- generate mode/register field from ea structure
 *
 *	return is a 6-bit field suitable for adding to a base
 *	opcode.
 */
int
modreg(register struct ea* e)
{
	register int mr = 0;

	switch(e->type){
	case DN:        mr = 000 + (int)e->reg; break;
	case AN:        mr = 010 + (int)e->reg; break;
	case ANI:       mr = 020 + (int)e->reg; break;
	case PSTINC:    mr = 030 + (int)e->reg; break;
	case PREDEC:    mr = 040 + (int)e->reg; break;
	case INDEX:     mr = e->itype == D16AN ? 050+(int)e->reg : 060+(int)e->reg; break;
	case IMMED:     mr = 074; break;
	case EXPR:      mr = e->siz==L ? 071 : 070; break;
	case PCINDEX:   mr = e->itype == D16AN ? 072 : 073; break;
	default:
		fatal("Bad type in modreg");
		}
	return(mr);
}

/*
 *      finish --- generate post-words for an instruction
 */
void
finish(struct ea* e)
{
	switch(e->type){
	case DN:
	case AN:
	case ANI:
	case PSTINC:
	case PREDEC:
		break;
	case PCINDEX:
	case INDEX:
		switch(e->itype){
		case D16AN:
			eword(check(e->constant,SWORD));
			break;
		case BRIEF:
			eword( genxreg(e) + check(e->constant,SBYTE));
			break;
		case FULL:
			if( e->xn_sup && e->prepst ) /* reserved combination */
				e->prepst = 0;
			eword( genxreg(e) + 0x100 + (e->br_sup<<7) +
				  (e->xn_sup<<6) + (e->bdsiz<<4) +
				  (e->prepst<<2) + e->odsiz );
			switch( e->bdsiz ){
			case 1:
				break;  /* supressed */
			case 2:
				eword(check(e->constant,SWORD));
				break;
			case 3:
				elong(e->constant);
				break;
			default:
				fatal("finish1");
				}
			switch( e->odsiz ){
			case 0:		/* to allow An indirect w/ index */
			case 1:
				break;  /* supressed */
			case 2:
				eword(check(e->const2,SWORD));
				break;
			case 3:
				elong(e->const2);
				break;
			default:
				fatal("finish2");
				}
			break;
		default:
			fatal("finish3");
			}
		break;
	case IMMED:
		if( Size == L )
			elong(e->constant);
		else
			eword(e->constant);        /* note: no range check here */
		break;
	case EXPR:
		if( e->siz==L )
			elong(e->constant);
		else
			eword(check(e->constant,UWORD));
		break;
	default:
		fatal("finish4");
		}
}

/*
 *      genxreg --- generate index register spec for indexed postword
 */
int
genxreg(struct ea* e)
{
	int     wrd;

	wrd = e->stat2 ? 0x8000 : 0x0000 ;
	wrd += (int)e->reg2<<12;
	if( e->siz == L )
		wrd += (1<<11);
	wrd += e->scl<<9;
	return(wrd);
}

/*
 *      getrlist --- return register list mask for an EA
 */
int
getrlist(struct ea *e)
{
	int rlist;

	switch(e->type){
	case AN:        rlist = 1<<((int)e->reg + 8);break;
	case DN:        rlist = 1<< (int)e->reg;	break;
	case RLIST:     rlist = (int)e->reg;		break;
	default:        fatal("getrlist");
		}
	return( rlist );
}

/*
 *      bitfld --- return bit field extension word
 */
WORD
bitfld(struct ea *e, int r)
{
	long     offset,width;

	if( e->type != FIELD )
		fatal("Botch in bitfld");

	if( e->stat )
		offset = check(e->constant,LOW5);
	else
		offset = e->reg + 0x20;

	if( e->stat2 )
		width = check(e->const2,LOW5);
	else
		width = e->reg2 + 0x20;

	return((WORD)( (r<<12) + (offset<<6) + width ));
}

char    *stypes[] = { "Symbol: ",  "Status:  ", "Data:   ", "Address: ", "Control: ",
			 "Float:  ",  "Fcontrol:", "PCrel:  ", "ZPCrel:  ", "ZAddress:",
			 "ZData:  ",  "Pmmu:    " };

/*
 *      dump_regs --- print reserved register names & classes
 */
void
dump_regs(void)
{
	register struct regs *r;
	int     lasttype = 999;

	printf("Reserved Register Names:");
	for(r=iregs; r<&iregs[NIREGS]; r++){
		if( isupper(r->r_name[0]))
			continue;       /* skip uppercase duplicate names */
		if( r->r_type != lasttype ){
			lasttype = r->r_type;
			printf("\n%s\t",stypes[lasttype]);
			}
		printf("%s ",r->r_name);
		}
	printf("\n\n");
}

/*
 *      do_opt --- scan operands and compare against list of options
 */
void
do_opt(struct opts *l)
{
	char    opt[100];
	char    arg[100];
	char    *p,*q;
	struct opts *lp;

	q = opt;
	arg[0] = '\0';
	while( (p=next_str()) != NULL ){
		while(*p){
			if( *p == '=' ){
				*q = '\0';
				q = arg;
				p++;
				}
			else
				*q++ = *p++;
			}
		*q = '\0';
		for(lp = l; lp->opt_name; lp++ ){
			if( strcmp(lp->opt_name,opt)==0){
				(*lp->opt_func)(arg);
				break;
				}
			}
		if(lp->opt_name)
			serror("Unrecognized OPT: %s",opt);
		}
}

/* opt functions */
void o_list(void){ Lflag=1; }
void o_nolist(void){ Lflag=0; }
void o_null(void){}


